import importlib
import logging
import os
import shutil
import tempfile
from abc import ABC
from typing import Dict, Type, BinaryIO, Any

from packaging import version  # type: ignore

from serialzy.api import Schema, VersionBoundary
from serialzy.base import DefaultSchemaSerializerByReference
from serialzy.utils import cached_installed_packages


# noinspection PyPackageRequirements
def serialize_to_file(dest: BinaryIO, save_to_file) -> None:
    with tempfile.NamedTemporaryFile() as handle:
        save_to_file(handle.name)
        while True:
            data = handle.read(8096)
            if not data:
                break
            dest.write(data)


def deserialize_from_file(source: BinaryIO, read_from_file) -> Any:
    with tempfile.NamedTemporaryFile() as handle:
        while True:
            data = source.read(8096)
            if not data:
                break
            handle.write(data)
        handle.flush()
        os.fsync(handle.fileno())
        return read_from_file(handle.name)


def serialize_to_dir(dest: BinaryIO, save_to_dir):
    tmpdir = tempfile.mkdtemp()
    save_to_dir(tmpdir)
    tar_save_path = os.path.join(tmpdir, "model")
    saved_archive = shutil.make_archive(tar_save_path, "tar", tmpdir)
    with open(saved_archive, "rb") as handle:
        while True:
            data = handle.read(8096)
            if not data:
                break
            dest.write(data)
    shutil.rmtree(tmpdir)


def deserialize_from_dir(source: BinaryIO, read_from_dir) -> Any:
    with tempfile.NamedTemporaryFile(suffix=".tar") as handle:
        while True:
            data = source.read(8096)
            if not data:
                break
            handle.write(data)
        handle.flush()
        os.fsync(handle.fileno())

        tmpdir = tempfile.mkdtemp()
        shutil.unpack_archive(handle.name, tmpdir, "tar")
        model = read_from_dir(tmpdir)
        shutil.rmtree(tmpdir)
        return model


class ModelBaseSerializer(DefaultSchemaSerializerByReference, ABC):
    def __init__(self, module: str, serializer_name: str):
        self.module = module
        self.logger = logging.getLogger(serializer_name)

    def available(self) -> bool:
        # noinspection PyBroadException
        try:
            importlib.import_module(self.module)

            return True
        except:
            return False

    def stable(self) -> bool:
        return True

    def meta(self) -> Dict[str, str]:
        m = importlib.import_module(self.module)

        return {self.module: m.__version__}

    def resolve(self, schema: Schema) -> Type:
        typ = super().resolve(schema)
        if self.module not in schema.meta:
            self.logger.warning(f'No {self.module} version in meta')
        elif version.parse(schema.meta[self.module]) > version.parse(cached_installed_packages[self.module]):
            self.logger.warning(f'Installed version of {self.module} {cached_installed_packages[self.module]} '
                                f'is older than used for serialization {schema.meta[self.module]}')
        return typ

    def requirements(self) -> Dict[str, VersionBoundary]:
        return {self.module: VersionBoundary()}



from typing import BinaryIO, Callable, Type, Union, Any, Optional

from packaging import version  # type: ignore

from serialzy.serializers.base_model import ModelBaseSerializer, serialize_to_file, deserialize_from_file


# noinspection PyPackageRequirements
class XGBoostSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("xgboost", __name__)

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        serialize_to_file(dest, lambda x: obj.save_model(x))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import xgboost  # type: ignore

        def load_model(filename):
            model = xgboost.Booster({'nthread': 4})
            model.load_model(filename)
            return model

        return deserialize_from_file(source, load_model)

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import xgboost as xgb  # type: ignore
        return lambda t: t in [xgb.XGBRanker, xgb.XGBRegressor, xgb.XGBClassifier]

    def data_format(self) -> str:
        return "xgb"



import inspect
from typing import Union, Type, Callable, Any, BinaryIO, Optional

from serialzy.serializers.base_model import ModelBaseSerializer, serialize_to_dir, deserialize_from_dir


# noinspection PyPackageRequirements
class TensorflowKerasSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("tensorflow", __name__)

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import tensorflow as tf  # type: ignore
        return lambda t: t in [tf.keras.Sequential] or (inspect.isclass(t) and issubclass(t, tf.keras.Model))

    def data_format(self) -> str:
        return "tf_keras"

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        serialize_to_dir(dest, lambda x: obj.save(x, save_format=self.data_format()))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import tensorflow as tf  # type: ignore
        return deserialize_from_dir(source, lambda x: tf.keras.models.load_model(x))


# noinspection PyPackageRequirements
class TensorflowPureSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("tensorflow", __name__)

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import tensorflow as tf  # type: ignore
        return lambda t: t in [tf.train.Checkpoint] or (inspect.isclass(t) and issubclass(t, tf.Module))

    def data_format(self) -> str:
        return "tf_pure"

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        import tensorflow as tf  # type: ignore
        serialize_to_dir(dest, lambda x: tf.saved_model.save(obj, x))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import tensorflow as tf  # type: ignore
        return deserialize_from_dir(source, lambda x: tf.saved_model.load(x))



from typing import BinaryIO, Callable, Type, Union, Any, Optional

from packaging import version  # type: ignore

from serialzy.serializers.base_model import ModelBaseSerializer, serialize_to_file, deserialize_from_file


# noinspection PyPackageRequirements
class LightGBMSerializer(ModelBaseSerializer):
    def __init__(self):
        super().__init__("lightgbm", __name__)

    def _serialize(self, obj: Any, dest: BinaryIO) -> None:
        serialize_to_file(dest, lambda x: obj.booster_.save_model(x))

    def _deserialize(self, source: BinaryIO, schema_type: Type, user_type: Optional[Type] = None) -> Any:
        self._check_types_valid(schema_type, user_type)
        import lightgbm  # type: ignore
        return deserialize_from_file(source, lambda x: lightgbm.Booster(model_file=x))

    def supported_types(self) -> Union[Type, Callable[[Type], bool]]:
        import lightgbm as lgbm  # type: ignore
        return lambda t: t in [lgbm.LGBMRanker, lgbm.LGBMRegressor, lgbm.LGBMClassifier]

    def data_format(self) -> str:
        return "lgbm"
